package org.jabber.applet.connection;

import java.net.*;
import java.io.*;
import java.util.*;
/***************************************************************************
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in
 * either source code or executable form, except in compliance with the 
 * License.  You may obtain a copy of the License at 
 * 
 * http://www.jabber.com/license/ 
 * 
 * or at
 * 
 * http://www.opensource.org/.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the
 * License for the specific language governing rights and limitations 
 * under the License.
 * 
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 ***************************************************************************/

/**
 * This class implements Jabber HTTP polling as defined in JEP-0025.
 * Jabber polling is used to tunnel the Jabber protocol through a http proxy.
 * 
 * @author	Matthias Wimmer
 * @version 	$ 0.7.1 $
 */
public class PollingConnection extends Connection {
    private boolean keepPolling = true;
    private int pollingInterval = 2500;
    private final int minPollingInterval = 2500;
    private final int maxPollingInterval = 30000;
    private URL pollingURL;
    private String sessionCookie;

    /**
     * constructor for the PollingConnection class
     *
     * @param connectionClient Reference to the IConnectionClient object
     * @param pollingURL The URL that has to be polled
     */
    public PollingConnection (IConnectionClient connectionClient, URL pollingURL) {
        super(connectionClient);
        this.pollingURL = pollingURL;
    }

    /**
     * send a polling request to the server and send data to it
     *
     * @param send data to be sent
     */
    private synchronized void poll(String send) {
        try {
            // get a HttpURLConnection and check if it is a HTTP URL
            URLConnection urlTemp = pollingURL.openConnection();
            if (!(urlTemp instanceof HttpURLConnection)) {
                System.out.println("URL ist not a HTTP URL");
                disconnect();
                return;
            }
            HttpURLConnection urlConn = (HttpURLConnection)urlTemp;

            // we have to POST the data
            urlConn.setRequestMethod("POST");

            // we send output
            urlConn.setDoInput(true);

            // we want read the reply
            urlConn.setDoOutput(true);

            // pass caches
            urlConn.setUseCaches(false);

            // allow the user to enter passwords if necessary
            urlConn.setAllowUserInteraction(true);

            InputStream in;
            OutputStream out;

            // we POST with a MIME type that is not really correct
            urlConn.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");

            // output session cookie "0" for new connection
            out = urlConn.getOutputStream();
            if (sessionCookie == null) { // new session, no cookie yet
                out.write("0,".getBytes("UTF8"));
            } else { // existing session and cookie
                out.write((sessionCookie+",").getBytes("UTF8"));
            }

            // do we have to send real content?
            if (send != null) {
                out.write(send.getBytes("UTF8"));
            }

            // flush and close the output so that the receiver gets an EOF
            out.flush();
            out.close();

            in = urlConn.getInputStream();

            // read the new session cookie
            sessionCookie = urlConn.getHeaderField("Set-Cookie");
            if (sessionCookie != null) {
                if (sessionCookie.substring(0,3).equals("ID=")) {
                    sessionCookie = sessionCookie.substring(3);
                }
                int index = sessionCookie.indexOf(';');
                if (index>0) {
                    sessionCookie = sessionCookie.substring(0, index-1);
                }
            }

            // check HTTP response code (should be 200)
            if (urlConn.getResponseCode() != urlConn.HTTP_OK) {
                System.out.println("HTTP response code is: "+(new Integer(urlConn.getResponseCode())).toString());
                disconnect();
            }

            // is it an error reply?
            if (sessionCookie.endsWith(":0")) {
                System.out.println("Got error cookie: "+sessionCookie);
                disconnect();
            } else {
                System.out.println("Got session cookie: "+sessionCookie);
            }

            // read the input
            long bytesReceived = 0;
            int count;
            byte[] b = new byte[1024];
            while (-1 != (count = in.read(b, 0, 1024))) {
                if (count > 0) {
                    String resp = new String(b, 0, count, "UTF8");
                    bytesReceived += count;
                    dataReceived(resp);
                } else { // we have to wait a bit for the reply
                    try {
                        sleep(1000);
                    } catch (InterruptedException ie) {}
                }
            }
            in.close();

            // recalculate the polling interval, adjust by 10%
            int tenthOfInterval = pollingInterval/10;

            // if we got something real it should be bigger then 2 bytes
            // if we get 1 or 2 bytes it's maybe just a keep-alive space
            if (bytesReceived > 2) {
                pollingInterval -= tenthOfInterval;
            } else {
                pollingInterval += tenthOfInterval;
            }

            // check if the new interval is okay
            if (pollingInterval > maxPollingInterval) {
                pollingInterval = maxPollingInterval;
            }
            if (pollingInterval < minPollingInterval) {
                pollingInterval = minPollingInterval;
            }

        } catch (IOException ioe) {
            disconnect();
        }
    }

    /**
     * send a polling request to the server
     */
    private void poll() {
        poll(null);
    }

    /**
     * Look for new data packets from Server.
     */
    public void run() {
        while (keepPolling) {
            System.out.println("PollingLoop");
            if (moConnState.state() == moConnState.OPEN) {
                poll();
            }
            try {
                sleep(pollingInterval);
            } catch(InterruptedException e) { }
        }
    }

    /**
     * Sends data packets to Server
     * 
     * @param str text to send to server 
     * @return true if it was able to send data, false otherwise 
     */
    public boolean send(String str) {
        if (moConnState.state() != moConnState.OPEN) {
            System.out.println("ContinuousConnection.Send() failed: connection not open - state is: "+moConnState.toString());
            return false;
        }

        poll(str);
        return moConnState.state() == moConnState.OPEN;
    }

    /**
     * Open virtual connection to the server. We don't really but set the
     * state to connected and start the polling loop.
     * 
     * @param host Host Name or IP Address (Ignored for PollingConnection)
     * @param port Port to connect to (Ignored for PollingConnection)
     * @return the new connection state
     */
    public ConnectionState connect(String host, int port) {
        moConnState.state(ConnectionState.OPEN);

        return moConnState;
    }

    /**
     * Disconnect logical connection to the server
     */
    public void disconnect() {
        System.out.println("PollingConnection.disconnect() called");
        moConnState.state(moConnState.CLOSING);

        if (moConnectionClient != null) {
            moConnectionClient.connectionClosed();
        }

        moConnState.state(moConnState.CLOSED);
    }

    /**
     * Prepare the instance of PollingConnection to be removed.
     * This will stopp the polling loop.
     */
    public void cleanup() {
        disconnect();
        keepPolling = false;
    }
}
